@@ -1,6 +1,7 @@
 from __future__ import unicode_literals
 from __future__ import absolute_import
 
+import hashlib
 import json
 import logging
 import requests
@@ -14,7 +15,8 @@ from pypuppetdb.errors import (
 )
 from pypuppetdb.types import (
     Node, Fact, Resource,
-    Report, Event, Catalog
+    Report, Event, Catalog,
+    Inventory
 )
 from pypuppetdb.QueryBuilder import *
 
@@ -44,6 +46,9 @@ ENDPOINTS = {
     'fact-contents': 'pdb/query/v4/fact-contents',
     'edges': 'pdb/query/v4/edges',
     'pql': 'pdb/query/v4',
+    'inventory': 'pdb/query/v4/inventory',
+    'status': 'status/v1/services/puppetdb-status',
+    'cmd': 'pdb/cmd/v1'
 }
 
 PARAMETERS = {
@@ -55,6 +60,13 @@ PARAMETERS = {
     'server_time': 'server_time',
 }
 
+COMMAND_VERSION = {
+    "deactivate node": 3,
+    "replace catalog": 9,
+    "replace facts": 5,
+    "store report": 8
+}
+
 ERROR_STRINGS = {
     'timeout': 'Connection to PuppetDB timed out on',
     'refused': 'Could not reach PuppetDB on',
@@ -114,11 +126,14 @@ class BaseAPI(object):
             authentication
     :type password: :obj:`None` or :obj:`string`
 
+    :param token: (optional) The X-auth token to use for X-Authentication
+    :type token: :obj:`None` or :obj:`string`
+
     :raises: :class:`~pypuppetdb.errors.ImproperlyConfiguredError`
     """
     def __init__(self, host='localhost', port=8080, ssl_verify=True,
                  ssl_key=None, ssl_cert=None, timeout=10, protocol=None,
-                 url_path=None, username=None, password=None):
+                 url_path=None, username=None, password=None, token=None):
         """Initialises our BaseAPI object passing the parameters needed in
         order to be able to create the connection strings, set up SSL and
         timeouts and so forth."""
@@ -130,6 +145,7 @@ class BaseAPI(object):
         self.ssl_key = ssl_key
         self.ssl_cert = ssl_cert
         self.timeout = timeout
+        self.token = token
 
         # Standardise the URL path to a format similar to /puppetdb
         if url_path:
@@ -155,6 +171,9 @@ class BaseAPI(object):
             'accept-charset': 'utf-8'
         }
 
+        if self.token:
+            self._session.headers['X-Authentication'] = self.token
+
         if protocol is not None:
             protocol = protocol.lower()
             if protocol not in ['http', 'https']:
@@ -162,6 +181,8 @@ class BaseAPI(object):
             self.protocol = protocol
         elif self.ssl_key is not None and self.ssl_cert is not None:
             self.protocol = 'https'
+        elif self.token is not None:
+            self.protocol = 'https'
         else:
             self.protocol = 'http'
 
@@ -318,19 +339,25 @@ class BaseAPI(object):
         if not (payload):
             payload = None
 
+        if not self.token:
+            auth = (self.username, self.password)
+        else:
+            auth = None
+
         try:
             if request_method.upper() == 'GET':
                 r = self._session.get(url, params=payload,
                                       verify=self.ssl_verify,
                                       cert=(self.ssl_cert, self.ssl_key),
                                       timeout=self.timeout,
-                                      auth=(self.username, self.password))
+                                      auth=auth)
             elif request_method.upper() == 'POST':
-                r = self._session.post(url, params=payload,
+                r = self._session.post(url,
+                                       data=json.dumps(payload, default=str),
                                        verify=self.ssl_verify,
                                        cert=(self.ssl_cert, self.ssl_key),
                                        timeout=self.timeout,
-                                       auth=(self.username, self.password))
+                                       auth=auth)
             else:
                 log.error("Only GET or POST supported, {0} unsupported".format(
                           request_method))
@@ -367,13 +394,87 @@ class BaseAPI(object):
                                                      self.protocol.upper()))
             raise
 
+    def _cmd(self, command, payload):
+        """This method posts commands to PuppetDB. Provided a command and payload
+        it will fire a request at PuppetDB. If PuppetDB can be reached and
+        answers within the timeout we'll decode the response and give it back
+        or raise for the HTTP Status Code yesPuppetDB gave back.
+
+        :param command: The PuppetDB Command we want to execute.
+        :type command: :obj:`string`
+        :param command: The payload, in wire format, specific to the command.
+        :type path: :obj:`dict`
+
+        :raises: :class:`~pypuppetdb.errors.EmptyResponseError`
+
+        :returns: The decoded response from PuppetDB
+        :rtype: :obj:`dict` or :obj:`list`
+        """
+        log.debug('_cmd called with command: {0}, data: {1}'.format(
+            command, payload))
+
+        url = self._url('cmd')
+
+        if command not in COMMAND_VERSION:
+            log.error("Only {0} supported, {1} unsupported".format(
+                list(COMMAND_VERSION.keys()), command))
+            raise APIError
+
+        params = {
+            "command": command,
+            "version": COMMAND_VERSION[command],
+            "certname": payload['certname'],
+            "checksum": hashlib.sha1(str(payload)  # nosec
+                                     .encode('utf-8')).hexdigest()
+        }
+
+        if not self.token:
+            auth = (self.username, self.password)
+        else:
+            auth = None
+
+        try:
+            r = self._session.post(url,
+                                   params=params,
+                                   data=json.dumps(payload, default=str),
+                                   verify=self.ssl_verify,
+                                   cert=(self.ssl_cert, self.ssl_key),
+                                   timeout=self.timeout,
+                                   auth=auth)
+
+            r.raise_for_status()
+
+            json_body = r.json()
+            if json_body is not None:
+                return json_body
+            else:
+                del json_body
+                raise EmptyResponseError
+
+        except requests.exceptions.Timeout:
+            log.error("{0} {1}:{2} over {3}.".format(ERROR_STRINGS['timeout'],
+                                                     self.host, self.port,
+                                                     self.protocol.upper()))
+            raise
+        except requests.exceptions.ConnectionError:
+            log.error("{0} {1}:{2} over {3}.".format(ERROR_STRINGS['refused'],
+                                                     self.host, self.port,
+                                                     self.protocol.upper()))
+            raise
+        except requests.exceptions.HTTPError as err:
+            log.error("{0} {1}:{2} over {3}.".format(err.response.text,
+                                                     self.host, self.port,
+                                                     self.protocol.upper()))
+            raise
+
     # Method stubs
 
-    def nodes(self, unreported=2, with_status=False, **kwargs):
+    def nodes(self, unreported=2, with_status=False, with_event_numbers=True,
+              **kwargs):
         """Query for nodes by either name or query. If both aren't
         provided this will return a list of all nodes. This method
-        also fetches the nodes status and event counts of the latest
-        report from puppetdb.
+        also (optionally) fetches the nodes status and (optionally)
+        event counts of the latest report from puppetdb.
 
         :param with_status: (optional) include the node status in the\
                            returned nodes
@@ -381,6 +482,14 @@ class BaseAPI(object):
         :param unreported: (optional) amount of hours when a node gets
                            marked as unreported
         :type unreported: :obj:`None` or integer
+        :param with_event_numbers: (optional) include the exact number of\
+                           changed/unchanged/failed/noop events when\
+                           with_status is set to True. If set to False
+                           only "some" string is provided if there are
+                           resources with such status in the last report.
+                           This provides performance benefits as potentially
+                           slow event-counts query is omitted completely.
+        :type with_event_numbers: :bool:
         :param \*\*kwargs: The rest of the keyword arguments are passed
                            to the _query function
 
@@ -395,39 +504,50 @@ class BaseAPI(object):
         if type(nodes) == dict:
             nodes = [nodes, ]
 
-        if with_status:
+        if with_status and with_event_numbers:
             latest_events = self.event_counts(
                 query=EqualsOperator("latest_report?", True),
                 summarize_by='certname'
             )
 
         for node in nodes:
-            node['status'] = None
+            node['status_report'] = None
             node['events'] = None
 
             if with_status:
-                status = [s for s in latest_events
-                          if s['subject']['title'] == node['certname']]
+                if with_event_numbers:
+                    status = [s for s in latest_events
+                              if s['subject']['title'] == node['certname']]
 
-                try:
-                    if node['latest_report_noop']:
-                        node['status'] = 'noop'
-                    else:
-                        node['status'] = node['latest_report_status']
+                    try:
+                        node['status_report'] = node['latest_report_status']
 
-                    if status:
-                        node['events'] = status[0]
-                except KeyError:
-                    if status:
-                        node['events'] = status = status[0]
-                        if status['successes'] > 0:
-                            node['status'] = 'changed'
-                        if status['noops'] > 0:
-                            node['status'] = 'noop'
-                        if status['failures'] > 0:
-                            node['status'] = 'failed'
-                    else:
-                        node['status'] = 'unchanged'
+                        if status:
+                            node['events'] = status[0]
+                    except KeyError:
+                        if status:
+                            node['events'] = status = status[0]
+                            if status['successes'] > 0:
+                                node['status_report'] = 'changed'
+                            if status['noops'] > 0:
+                                node['status_report'] = 'noop'
+                            if status['failures'] > 0:
+                                node['status_report'] = 'failed'
+                        else:
+                            node['status_report'] = 'unchanged'
+                else:
+                    node['status_report'] = node['latest_report_status']
+                    node['events'] = {
+                        'successes': 0,
+                        'failures': 0,
+                        'noops': 0,
+                    }
+                    if node['status_report'] == 'changed':
+                        node['events']['successes'] = 'some'
+                    elif node['status_report'] == 'noop':
+                        node['events']['noops'] = 'some'
+                    elif node['status_report'] == 'failed':
+                        node['events']['failures'] = 'some'
 
                 # node report age
                 if node['report_timestamp'] is not None:
@@ -438,17 +558,17 @@ class BaseAPI(object):
                         unreported_border = now - timedelta(hours=unreported)
                         if last_report < unreported_border:
                             delta = (now - last_report)
-                            node['status'] = 'unreported'
+                            node['unreported'] = True
                             node['unreported_time'] = '{0}d {1}h {2}m'.format(
                                 delta.days,
                                 int(delta.seconds / 3600),
                                 int((delta.seconds % 3600) / 60)
                             )
                     except AttributeError:
-                        node['status'] = 'unreported'
+                        node['unreported'] = True
 
                 if not node['report_timestamp']:
-                    node['status'] = 'unreported'
+                    node['unreported'] = True
 
             yield Node(self,
                        name=node['certname'],
@@ -457,8 +577,11 @@ class BaseAPI(object):
                        report_timestamp=node['report_timestamp'],
                        catalog_timestamp=node['catalog_timestamp'],
                        facts_timestamp=node['facts_timestamp'],
-                       status=node['status'],
+                       status_report=node['status_report'],
+                       noop=node.get('latest_report_noop'),
+                       noop_pending=node.get('latest_report_noop_pending'),
                        events=node['events'],
+                       unreported=node.get('unreported'),
                        unreported_time=node.get('unreported_time'),
                        report_environment=node['report_environment'],
                        catalog_environment=node['catalog_environment'],
@@ -808,10 +931,43 @@ class BaseAPI(object):
                 transaction=report['transaction_uuid'],
                 environment=report['environment'],
                 status=report['status'],
-                noop=report['noop'],
+                noop=report.get('noop'),
+                noop_pending=report.get('noop_pending'),
                 metrics=report['metrics']['data'],
                 logs=report['logs']['data'],
                 code_id=report.get('code_id'),
                 catalog_uuid=report.get('catalog_uuid'),
                 cached_catalog_status=report.get('cached_catalog_status')
             )
+
+    def inventory(self, **kwargs):
+        """Get Node and Fact information with an alternative query syntax
+        for structured facts instead of using the facts, fact-contents and
+        factsets endpoints for many fact-related queries.
+
+        :param \*\*kwargs: The rest of the keyword arguments are passed
+                           to the _query function.
+
+        :returns: A generator yielding Inventory
+        :rtype: :class:`pypuppetdb.types.Inventory`
+        """
+        inventory = self._query('inventory', **kwargs)
+        for inv in inventory:
+            yield Inventory(
+                node=inv['certname'],
+                time=inv['timestamp'],
+                environment=inv['environment'],
+                facts=inv['facts'],
+                trusted=inv['trusted']
+            )
+
+    def status(self):
+        """Get PuppetDB server status.
+
+        :returns: A dict with the PuppetDB status information
+        :rtype: :obj:`dict`
+        """
+        return self._query('status')
+
+    def command(self, command, payload):
+        return self._cmd(command, payload)
